/*****************************************************************************

Compiling Method Call's to Atomic Safe Versions

  Because of the complication with the hierarchy of objects compiling method
calls to their atomic safe versions becomes a non-trival task. 

______________________________________________________________________________
Arrays & Objects with static type Object:

[Explanation]
  These can not be changed to AObject, only instances where we see 
"new Object" is it safe to translate to AObject. Therefore, if there exists
a call where the receiver has static type Object then translating the
non-atomic method name to an atomic one will fail at javac time (even though
this might be an AObject and have that method). This will always be a
problem with Arrays as well since the only methods you could call on an array
are the methods you would call on an object with static type Object. 

[Solution]
  To handle this we convert all calls that have an Array or Object type to
go through the wrappers in polyglot.ext.atomjava.runtime.java.lang.Object.
Within each wrapper we do instanceof/getClass().isArray() in order to
determine the runtime type. If the runtime type is unknown there exists
concern about what might happen if a non-atomic version is called (jump 
into a new atomic context, etc).
------------------------------------------------------------------------------











*****************************************************************************/
package polyglot.ext.atomjava.ast;

import polyglot.ext.jl.ast.Call_c;
import polyglot.visit.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;
import polyglot.ast.*;

import polyglot.ext.atomjava.types.*;
import polyglot.ext.atomjava.visit.*;
import polyglot.ext.atomjava.util.*;

import polyglot.ext.jl.types.*;

import java.util.*;

public class AtomJavaCall_c extends Call_c implements AtomJavaBuild {

  public AtomJavaCall_c(Position pos, Receiver target, String name, List args) {
    super(pos, target, name, args);      
  }
  
  public NodeVisitor buildEnter(AtomJavaBuilder ab) {
    return ab;
  }

  public Node build(Node parent, AtomJavaBuilder ab) {
    if(ab.context().inAtomicContext())
      return buildAtomic(parent, ab);
    return buildNotAtomic(parent, ab);
  }

  /* if this method throws checcked exceptions create a new method to handle this */
  private Node makeFailure(Node parent, AtomJavaBuilder ab) {
    AtomJavaContext     c  = ab.context();
    AtomJavaTypeSystem  ts = ab.typeSystem();
    AtomJavaNodeFactory nf = ab.nodeFactory();

    /* the target and args need to be evaluated before
       the call to failure (because they themselves might avert
       a failure), but we can't evaluate 'super' (by itself)
       and who cares about evaluating 'this' (both instanceof Special)
     */
    Expr e = 
	target() instanceof Expr && 
	!(target() instanceof Special)
	? (Expr) target() : nf.NullLit(c);

    ab.errorQueue().enqueue(ErrorInfo.WARNING, 
			    "No wrapper method found for method " + 
			    mi.signature() + " in " + mi.container() + 
			    ".", position());

    Call fail = nf.Failure(c, type(), toString(), 
                           AJS.cons(e, new LinkedList(arguments())));

    if(c.inTryCatchContext()) {

      List throwTypes = new LinkedList();
      for(Iterator i = methodInstance().throwTypes().iterator(); 
          i.hasNext();) {
        Type t = (Type) i.next();
        if(!ts.isUncheckedException(t))
          throwTypes.add(nf.CanonicalTypeNode(c, t));
      }

      if(throwTypes.size() > 0)
        fail = nf.Call(c, null, AJS.cons(e, new LinkedList(arguments())),
                       c.makeThrowingFailureMethod(this, throwTypes, ab));
    }

    /* cast calls where the object is not lost */
    if(ts.isSubtype(type(), ts.Object()) && !(parent instanceof Eval))
      return nf.Cast(c, nf.CanonicalTypeNode(c, type()), fail);
    return fail;
  }

  // for things in java.lang etc.
  private Node getWrapper(ReferenceType tt, MethodInstance mi,
                          Node parent, AtomJavaBuilder ab) {
    AtomJavaContext     c  = ab.context();
    AtomJavaTypeSystem  ts = ab.typeSystem();
    AtomJavaNodeFactory nf = ab.nodeFactory();

    try {
      ClassType wt = (ClassType) 
        ts.typeForName(AtomJavaTypeSystem.RUNTIME + "wrapper." +
                       ((ClassType)tt).fullName());
      
      MethodInstance m = 
        new MethodInstance_c(ts, Position.COMPILER_GENERATED, wt, 
                             mi.flags(), mi.returnType(), mi.name(), 
                             AJS.cons(tt, new LinkedList(mi.formalTypes())), 
                             Collections.EMPTY_LIST);
      
      /* look for wrapper */
      ts.findMethod(wt, m.name(), m.formalTypes(), c.currentClass()); 

      Expr target = target() instanceof Expr 
        ? (Expr) target() : nf.NullLit(c);
      
      Call wrapper = 
        nf.Call(c, nf.CanonicalTypeNode(c, wt), name(), 
                AJS.cons(target, new LinkedList(arguments())));

      return wrapper;
    } catch (SemanticException e) {}
  
    return makeFailure(parent,ab);
  }

  public Node buildAtomic(Node parent, AtomJavaBuilder ab) {
    AtomJavaContext     c  = ab.context();
    AtomJavaTypeSystem  ts = ab.typeSystem();
    AtomJavaNodeFactory nf = ab.nodeFactory();

    Call call = name(c.atomicMethodName(name));
    MethodInstance mi = call.methodInstance();
    mi = mi.name(call.name());
    call = (Call) call.methodInstance(mi);

    // FIX: Dan changed this to wrap only for java.lang etc. and fail 
    //      if otherwise unavailable (may want to expand)
//    Type t = call.target().type();
    Type t = mi.container();
    if(t instanceof ClassType && ts.useWrappers((ClassType)t))
      return ((AtomJavaCall_c)call).getWrapper((ClassType)t, mi, parent, ab);
    try { 
      ts.findMethod((ClassType)t, mi.name(), 
                    mi.formalTypes(),c.currentClass()); 
      return call;
    } catch(SemanticException e) {}
    /* otherwise, make something that fails at run-time */
    return ((AtomJavaCall_c)call).makeFailure(parent,ab);
  }

  public Node buildNotAtomic(Node parent, AtomJavaBuilder ab) {
      Type t = methodInstance().container();
//      Type t = target().type();
      if(t instanceof ClassType && ab.typeSystem().useWrappers((ClassType)t))
      return getWrapper((ClassType)t, methodInstance(), parent, ab);
    return this;
  }
}
